' ================================================================
' = mailmerger1.vbs (C) 2022 Pearbright Ltd all rights reserved  =
' ================================================================
' = set this non-zero to have more diagnostic information        =
' ================================================================
Const Verbosity = 0

Const ForReading = 1
Const ForWriting = 2

Const adOpenStatic = 3
Const adLockOptimistic = 3

Public token_array(1000)
Public nTokens
Public mapping_left(200)
Public mapping_right(200)
Public mapping_columns(200)
Public access_data(200)
Public nMappings
Public col_headers(1000)
Public nColumns

Function lookup(a)
' ================================================================
' = find a column header                                         =
' ================================================================
   Dim i, r
   r = ""
   i = 0
   While i < nColumns
       If col_headers(i) = a Then
           r = Cstr(i)
           i = nColumns
       End If
       i = i + 1
   Wend
   lookup = r
End Function

Function mapcolumns()
' ================================================================
' = index the column headers to avoid repeatedly searching array =
' ================================================================
    Dim i, j, k, a, s, r
    r = 0
    i = 0
    While i <= nMappings
       s = ""
       a = mapping_right(i)
       x = Split(a, ":")
       j = 0
       While j <= Ubound(x)
          y = Split(x(j), "+")
          k = 0
          While k <= Ubound(y)
              a = Trim(y(k))
              If Left(a & " ",1) <> "{" Then
                  a = lookup(a)
              End If
              If a = "" Then
                 r = 1
              End If
              s = s & a
              If k < Ubound(y) Then
                  s = s & "+"
              End If
              k = k + 1
          Wend
          If j < Ubound(x) Then
              s = s & ":"
          End If
          j = j + 1
       Wend
       mapping_right(i) = s
       i = i + 1
    Wend
    mapcolumns = r
End Function

Function isinitial(a)
' ================================================================
' = tests a string to see if it should be rejected as unsuitable =
' = for salutation e.g.                                          =
' = John Paul Jones -> Dear John                                 =
' = T S Eliot -> Dear Mr Eliot                                   =
' = Mr X -> Dear Friend                                          =
' ================================================================
   Dim b, r
   r = 0
   b = Trim(a)
   If Len(b) > 1 Then
      r = 1
      If Mid(b, 2, 1) = " " Then
         r = 0
      End If
   End If
   isinitial = r
End Function

Function mapdata2(a, z)
' ================================================================
' = concatenate multiple fields if mapping contains plus sign;    =
' = if token isn't a literal e.g. {Supporter} then use it to     =
' = select the appropriate field from the input data file        =
' ================================================================
    Dim s, i, b
    If Left(a & " ", 1) = "{" Then
       If Len(a) > 2 Then
          s = Mid(a, 2, Len(a) - 2)
       Else
          s = ""
       End If
    Else
       If 0 < Instr(a, "+") Then
           s = ""
           x = Split(a, "+")
           i = 0
           While i <= Ubound(x)
              b = Trim(token_array(Cint(Trim(x(i)))))
              If z <> 0 Then
' ================================================================
' = for salutation type fields reject blanks or initials         =
' ================================================================
                  If isinitial(b) = 0 Then
                     s = ""
                     b = ""
                     i = 999
                  End If
              End If
              s = Trim(s & " " & b)
              i = i + 1
           Wend
       Else
           Err.Clear
           On Error Resume Next
           s = Trim(token_array(Cint(a)))
           If Err.Number <> 0 Then
               If Verbosity <> 0 Then
                   s = "Error: " & Trim(Cstr(Err.Number)) & " [" & Err.Description & "] {" & a & "}"
                   Wscript.Echo s
               End If
               s = ""
           End If
           On Error Goto 0
' ================================================================
' = for salutation type fields reject blanks or initials         =
' ================================================================
           If z <> 0 And 0 = isinitial(s) Then
               s = ""
           End If
       End If
    End If
    mapdata2 = s
End Function

Function mapdata()
' ================================================================
' = map a row of data from the incoming CSV file into MS Access  =
' ================================================================
    Dim i, a, s, j, u, w
    i = 0
    While i <= nMappings
        a = mapping_right(i)
        If 0 < Instr(a, ":") Then
' ================================================================
' = If separated by colons apply alternatives for salutation     =
' ================================================================
           x = Split(a, ":")
           j = 0
           While j <= Ubound(x)
              s = Trim(x(j))
              u = 0
              If Left(s & " ", 1) = "{" Then
                 u = 1
              End If
              s = mapdata2(s, 1)
              If s <> "" Then
                 access_data(i) = s
                 j = 999
              End If
              j = j + 1
           Wend
        Else
           access_data(i) = mapdata2(a, 0)
        End If
        i = i + 1
    Wend
    mapdata = 0
End Function

Function requote(a)
' ================================================================
' = put quotes round a string, and escape any quotes within it   =
' ================================================================
    Dim b, i, c, q
    q = "'"
    b = q
    i = 0
    While i < Len(a)
        i = i + 1
        c = Mid(a, i, 1)
        If c = q Then
            b = b & q & q
        Else
            b = b & c
        End If
    Wend
    b = b & q
    requote = b
End Function

Function GetMappings(f)
' ================================================================
' = read the mapping file, which defines the relationship        =
' = between the incoming data and the fields in MS Access        =
' ================================================================
   Dim textLine, i
   nMappings = -1
   Set objFSO = CreateObject("Scripting.FileSystemObject")
   If objFSO.FileExists(f) Then
       Set objIpFile = objFSO.OpenTextFile(f, ForReading)
       Do Until objIpFile.AtEndofStream
          textLine = objIpFile.Readline
          i = tokenize(textLine)
          If i = 2 And nMappings < 200 Then
              nMappings = nMappings + 1
              mapping_left(nMappings) = token_array(0)
              mapping_right(nMappings) = token_array(1)
          End If
       Loop
       objIpFile.Close
   Else
       Wscript.Echo("File [" & f & "] not found")
   End If
   GetMappings = nMappings + 1
End Function

Function MenuSelect(sTitle, arrPrompt, iMaxnum)
' ================================================================
' = display a menu and obtain response from user                 =
' ================================================================
 Dim n, j, s, r, sCrLf
 r = -2
 n = Ubound(arrPrompt)
 If n > iMaxnum Then
  n = iMaxnum
 End If
 s = ""
 j = 0
 sCrLf = Chr(13) & Chr(10)
 While j <= n
  s = s & Right("   " & Trim(Cstr(j+1)) & ") ", 5) & arrPrompt(j) & sCrLf
  j = j + 1
 Wend
 s = s & "  0) Cancel" & sCrLf
 While r < -1 Or r > n
' ================================================================
' = we could use InputBox like so: s = InputBox(s, sTitle, "0")  =
' = but instead we use the console so the font size can be       =
' = increased by calling the BAT file from a shortcut            =
' ================================================================
  WScript.StdOut.Write(sCrLf)
  WScript.StdOut.Write(sCrLf)
  WScript.StdOut.Write(sCrLf)
  WScript.StdOut.Write(sTitle)
  WScript.StdOut.Write(sCrLf)
  WScript.StdOut.Write(sCrLf)
  WScript.StdOut.Write(s)
  WScript.StdOut.Write("? ")
  WScript.StdIn.Read(0)
  s = WScript.StdIn.ReadLine()
  WScript.StdOut.Write(sCrLf)
  If IsNumeric(s) Then
     r = Cint(s) - 1
  End If
 Wend
 s = ""
 If r > -1 And r <= n Then
  s = arrPrompt(r)
 End If
 MenuSelect = s
End Function

Function DockTail(s)
' ================================================================
' = remove the extension from a filename                         =
' ================================================================
 Dim a, k
 k = InStrRev(s,".")
 a = ""
 If k > 1 Then
  a = Left(s, k - 1)
 End If
 Docktail = a
End Function

Function dequote(a)
' ================================================================
' = remove the quotes from a quoted string                       =
' ================================================================
        Dim b, q, qq, n
        q = """"
        qq = q & q
        b = a
        If Left(b, 1) = q Then
            If b = q Then
                b = ""
            Else
                b = Mid(b, 2)
                If Right(b, 1) = q Then
                    If b = q Then
                        b = ""
                    Else
                        b = Left(b, Len(b) - 1)
                        n = 0
                        While n < Len(b)
                            n = n + 1
                            If Mid(b, n, 2) = qq Then
                                b = Left(b, n - 1) & Mid(b, n + 1)
                            End If
                        Wend
                    End If
                End If
            End If
        End If
        dequote = b
End Function

Function findnextcomma(a)
' ================================================================
' = find the next comma that's not within quotes                 =
' ================================================================
        Dim q, r, z, c
        q = 1
        z = 0
        r = 0
        While q <= Len(a)
            c = Mid(a, q, 1)
            q = q + 1
            If c = """" Then
                If z = 0 Then
                    z = 1
                Else
                    z = 0
                    If q <= Len(a) Then
                        c = Mid(a, q, 1)
                        If c = """" Then
                            z = 1
                            q = q + 1
                        End If
                    End If
                End If
            Else
                If c = "," Then
                    If z = 0 Then
                        r = q - 1
                        q = Len(a) + 1
                    End If
                End If
            End If
        Wend
        findnextcomma = r
End Function

Function tokenize(t)
' ================================================================
' = split up a comma-separated string of tokens                  =
' ================================================================
        Dim p, n, z, a, token_string
        token_string = t
        a = ""
        p = 0
        n = 0
        z = 1
' ================================================================
' = clear the tokenizing array                                   =
' ================================================================
        n = 0
        While n < nTokens
            token_array(n) = ""
            n = n + 1
        Wend
        n = 0
' ================================================================
        While z <> 0 And n < 1000
            z = findnextcomma(token_string)
            If z > 0 Then
                z = z - 1
                If z < 1 Then
                    a = ""
                Else
                    a = Left(token_string, z)
                End If
                token_array(n) = Trim(dequote(a))
                n = n + 1
                z = z + 2
                If z > Len(token_string) Then
                    token_string = ""
                Else
                    token_string = Mid(token_string, z)
                End If
            Else
                If token_string <> "" Then
                    token_array(n) = Trim(dequote(token_string))
                    n = n + 1
                End If
            End If
        Wend
        nTokens = n ' so we know how much to clear next time
        tokenize = n
End Function

Function myformat(i)
' ================================================================
' = for formatting the date and time                             =
' ================================================================
   Dim s
   s = "0000" & Trim(Cstr(i))
   If i > 99 Then
       s = Right(s,4)
   Else
       s = Right(s,2)
   End If
   myformat = s
End Function

' ================================================================
' = START OF MAIN PROGRAM                                        =
' ================================================================
Dim sHomeFolder, sTemplateFolder, sFileName, i, s, iMaxnum, sCrLf
Dim sSelectedTemplate, sMappingFile, sLogFile, n, f, dt, sQuot
' ================================================================
' = maximum number of templates is 41                            =
' ================================================================
Dim nTemplates
Dim sTemplates(40)
nTemplates = 40
' ================================================================
sCrLf = Chr(13) & Chr(10)
sQuot = """"
' ================================================================
' = clear the tokenizing array                                   =
' ================================================================
n = 0
While n < 1000
    token_array(n) = ""
    n = n + 1
Wend
n = 0
nTokens = 1
' ================================================================
' = get the command line arguments, there should be one          =
' ================================================================
Set oArgs = WScript.Arguments
If oArgs.Count <> 1 Then
 WScript.StdOut.Write("Usage is CSCRIPT mailmerge1.vbs //Nologo <path of working folder>" & sCrLf)
Else
 sHomeFolder = oArgs.Item(0)
 Set objFSO = CreateObject("Scripting.FileSystemObject")
 Set objFolder = objFSO.GetFolder(sHomeFolder)
 s = objFolder.Path
 WScript.StdOut.Write("Working directory is [" &s & "]" & sCrLf)
 sHomeFolder = s
 sTemplateFolder = s & "\templates"
 sMappingFolder = s & "\mapping"
 sLogFile = s & "\mailmerger1.log"
' ================================================================
' Open a log file                                                =
' ================================================================
 If Not objFSO.FileExists(sLogFile) Then
      Set objLogFile = objFSO.CreateTextFile(sLogFile)
      objLogFile.Close
 End If 
 Set objLogFile = objFSO.OpenTextFile(sLogFile, ForWriting, True)
 dt=now
 s = "Started at " & myformat(hour(dt)) & ":" & myformat(minute(dt)) & " on " & myformat(day(dt)) & "/" & myformat(month(dt)) & "/" & myformat(year(dt))
 objLogFile.WriteLine(s)
 objLogFile.WriteLine("Templates should be in [" & sTemplateFolder & "]")
 objLogFile.WriteLine("Mappings should be in [" & sMappingFolder & "]")

' ================================================================
' = Loop through the files in the templates folder               =
' ================================================================
 Set objFolder = objFSO.GetFolder(sTemplateFolder)
 i = 0
 For Each objFile In objFolder.Files
  sTemplates(i) = objFile.Name
  iMaxnum = i
  i = i + 1
  If i > 40 Then
   WScript.StdOut.Write("There are too many templates in the template folder" & sCrLf)
   i = 0
  End If
 Next
' ================================================================
' = Ask the user to pick one                                     =
' ================================================================
 s = MenuSelect("Please choose your template", sTemplates, iMaxnum)
 If s <> "" Then
' ================================================================
' = user has picked one, is there a mapping file?                =
' ================================================================
  sSelectedTemplate = s
  s = sMappingFolder & "\" & DockTail(s) & ".csv"
  If objFSO.FileExists(s) Then
   sMappingFile = s
   objLogFile.WriteLine("Found corresponding mapping file [" & sMappingFile & "]")
  Else
   sMappingFile = sMappingFolder & "\default.csv"
   objLogFile.WriteLine("Using default mapping file [" & sMappingFile & "]")
  End If
  i = GetMappings(sMappingFile)
  i = 0
  While i <= nMappings
      objLogFile.WriteLine("[" & mapping_left(i) & "] -> [" & mapping_right(i) & "]")
      i = i + 1
  Wend
' ================================================================
' = Open a connection to the MS-Access database                  =
' ================================================================
  s = sHomeFolder & "\address.mdb"
  WScript.StdOut.Write("Connecting to [" & s & "]" & sCrLf)
  objLogFile.WriteLine("Connecting to [" & s & "]")
  Set objConnection = CreateObject("ADODB.Connection")
  objConnection.Open "Driver={Microsoft Access Driver (*.mdb, *.accdb)};DBQ=" & s
' ================================================================
' = create the table to hold addresses                           =
' ================================================================
  s = "CREATE TABLE Office_Address_List ("
  i = 0
  While i <= nMappings
      If i > 0 Then
         s = s & ","
      End If
      s = s & " [" & mapping_left(i) & "] VARCHAR(40)"
      i = i + 1
  Wend
  s = s & ");"
  objLogFile.WriteLine(s)
  objConnection.Execute s
' ================================================================
' = now read the input data file                                 =
' ================================================================
   f = sHomeFolder & "\address.csv"
   If objFSO.FileExists(f) Then
       Set objIpFile = objFSO.OpenTextFile(f, ForReading)
       WScript.StdOut.Write("Reading " & f & sCrLf)
       objLogFile.WriteLine("Reading " & f)
       n = 0
       Do Until objIpFile.AtEndofStream
          s = objIpFile.Readline
          i = tokenize(s)
          If n = 0 Then
              nColumns = nTokens
              objLogFile.WriteLine("Data file has " & Cstr(nColumns) & " columns")
              i = 0
              While i < nTokens
                  col_headers(i) = token_array(i)
                  i = i + 1
              Wend
              i = mapcolumns()
              If i <> 0 Then
                  s = "Warning: there are columns in the mapping file which do not match the data file."
                  WScript.StdOut.Write(s & sCrLf)
                  objLogFile.WriteLine(s)
              End If
              i = 0
              While i <= nMappings
                  objLogFile.WriteLine("[" & mapping_left(i) & "] -> [" & mapping_right(i) & "]")
                  i = i + 1
              Wend
          Else
' ================================================================
' = map the input data to the access table                       =
' ================================================================
              i = mapdata()
              s = "INSERT INTO `Office_Address_List` VALUES ( "
              i = 0
              While i <= nMappings
                 If i > 0 Then
                    s = s & ","
                 End If
                 s = s & requote(access_data(i))
                 i = i + 1
              Wend
              s = s & " );"
' ================================================================
' = if verbosity not turned on, only log first 20 INSERT's       =
' ================================================================
              If Verbosity <> 0 Or n < 20 Then
                 objLogFile.WriteLine(s)
              End If
              If Verbosity = 0 And n = 20 Then
                 objLogFile.WriteLine("...")
              End If
' ================================================================
' = execute the INSERT                                           =
' ================================================================
              objConnection.Execute s
          End If
          n = n + 1
       Loop
       objIpFile.Close
       Set objIpFile = Nothing
' ================================================================
' = now open Word to do the mailmerge                            =
' ================================================================
       Set oWord = CreateObject("Word.Application")
       oWord.Visible = False
       s = sQuot & sTemplateFolder & "\" & sSelectedTemplate & sQuot
       objLogFile.WriteLine("Opening [" & s & "] in Word")
       Set oDoc = oWord.Documents.Add(s)
       oDoc.Activate
       Set oDoc = oWord.ActiveDocument
       s = oDoc.Name
       objLogFile.WriteLine("Active Word document is [" & s & "]")
       Set oOriginalDoc = oDoc
       Set oMailMerge = oDoc.MailMerge
       s = sHomeFolder & "\address.mdb"
       oMailMerge.OpenDataSource s
       s = oMailMerge.DataSource.Name
       objLogFile.WriteLine("Datasource is [" & s & "]")
' ================================================================
' = execute the mailmerge and save the result
' ================================================================
       'checking value of named constant: wdSendToNewDocument
       oMailMerge.Destination = 0
       oMailMerge.SuppressBlankLines = True
       oMailMerge.DataSource.QueryString = "SELECT * FROM `Office_Address_List`"
       'setting to values of named constants: wdDefaultFirstRecord and wdDefaultLastRecord
       oMailMerge.DataSource.FirstRecord = 1
       oMailMerge.DataSource.LastRecord = -16
       'checking value of named constant: wdMainAndDataSource
       If oMailMerge.State = 2 Then
           oMailMerge.Execute
' ================================================================
' = this should make a new document active                       =
' ================================================================
           Set oDoc = oWord.ActiveDocument
           s = oDoc.Name
           objLogFile.WriteLine("New active Word document is [" & s & "]")
           s = sHomeFolder & "\result.docx"
           objLogFile.WriteLine("Saving as [" & s & "]")
           oDoc.SaveAs s
           oDoc.Close 0
       End If
       oOriginalDoc.Close 0
       ' wdDoNotSaveChanges = 0
       oWord.Quit 0
       Set oMailMerge = Nothing
       Set oDoc = Nothing
       Set oWord = Nothing
   Else
       s = "File [" & f & "] not found"
       WScript.StdOut.Write(s & sCrLf)
       objLogFile.WriteLine(s)
   End If
' ================================================================
' = end of input data file                                       =
' ================================================================
  objConnection.Close
  Set objConnection = Nothing
' ================================================================
' = End If user picked a template                                =
' ================================================================
 End If
 objLogFile.Close
 Set objLogFile = Nothing
 Set objFile = Nothing
 Set objFolder = Nothing
 Set objFSO = Nothing
' ================================================================
' = End If correct number of command line arguments              =
' ================================================================
End If
